from pygrandchem.grandchem import StaggeredKernelParams

from pystencils.simp.liveness_opts import *
from pystencils.simp.liveness_opts_exp import *
import random
import pycuda.driver as drv

import pystencils as ps

from pystencils import show_code
from timeit import default_timer as timer

import copy

optSequenceCache = {}

all_opts = [[atomize_eqs, []], [schedule_eqs, [2]], [duplicate_trivial_ops, [3, 1]],
            [merge_field_accesses, []], [refuse_eqs, [1, 1]], [var_to_shmem, [4]],
            [var_to_shmem_lt, [4]]]


def mutateOptSequence(seq):
    changed = False
    new_seq = copy.deepcopy(seq)
    while not changed:
        choice = random.randint(0, 4)
        if choice == 0:
            new_seq.opts.append(random.choice(all_opts))
            changed = True
        elif choice == 1:
            if len(new_seq.opts) > 1:
                a = random.randint(0, len(new_seq.opts) - 1)
                b = random.randint(0, len(new_seq.opts) - 1)
                new_seq.opts[a], new_seq.opts[b] = new_seq.opts[b], new_seq.opts[a]
                changed = True
        elif choice == 2:
            if len(new_seq.opts) > 0:
                new_seq.opts.remove(random.choice(new_seq.opts))
                changed = True
        elif choice == 3:
            if len(new_seq.opts) > 0:
                opt = random.choice(new_seq.opts)
                change = random.choice([-1, 1])
                factor = 1
                if change < 0:
                    factor = random.uniform(0.3, 1.0)
                if change > 0:
                    factor = random.uniform(1.0, 3.0)

                if len(opt[1]) > 0:
                    arg = random.randint(0, len(opt[1]) - 1)

                    opt[1][arg] = int(max(0, opt[1][arg] * factor + change))
                    changed = True
        else:
            dim = random.randint(0, 2)
            change = random.randint(0, 1)
            newBlockSize = list(seq.blockSize)

            if change == 0:
                newBlockSize[dim] = min(512, newBlockSize[dim] * 2)
            else:
                newBlockSize[dim] = max(1, newBlockSize[dim] // 2)

            if newBlockSize[0] * newBlockSize[1] * newBlockSize[2] <= 512 and (
                    newBlockSize[0] >= 32 or newBlockSize[0] >= seq.blockSize[0]):
                seq.blockSize = tuple(newBlockSize)
                changed = True

    return new_seq


def evolvePopulation(pop, eqs_set, dhs, staggered_params=None):

    pop.append(livenessOptSequence())
    once_mutated = [mutateOptSequence(seq) for seq in pop[0:6]]
    twice_mutated = [mutateOptSequence(mutateOptSequence(seq)) for seq in pop[0:4]]
    thrice_mutated = [
        mutateOptSequence(mutateOptSequence(mutateOptSequence(seq))) for seq in pop[0:3]
    ]

    new_pop = list(set(pop + once_mutated + twice_mutated + thrice_mutated))

    scores = []
    for seq in new_pop:
        scores.append((seq, *rateSequence(seq, eqs_set, dhs, staggered_params)))

    old_scores = []
    for s in optSequenceCache:
        if s not in new_pop:
            if s not in optSequenceCache:
                print("Not in optSequenceCache: ")
                print(s)
                print(hash(s))
            old_scores.append((s, optSequenceCache[s][0], [0, 0]))
    old_scores.sort(key=lambda s: sum(s[1]))

    if len(old_scores) > 0: scores.append(old_scores[0])

    print()
    scores.sort(key=lambda s: sum(s[1]))
    new_pop = []
    count_old_seqs = 0

    for score in scores:
        if score[0] not in optSequenceCache:
            print("Everything in scores: ")
            for s in scores:
                print(s[0])

            print("Not in optSequenceCache: ")
            print(score[0])
            print(hash(score[0]))

        survive = False
        if (len(new_pop) < 4 or count_old_seqs < 3) and len(new_pop) < 10:
            if optSequenceCache[score[0]][1] > 3:
                count_old_seqs += 1
            new_pop.append(score[0])
            survive = True

        print("".join(["{:6.2f} ".format(sc) for sc in score[1]]) + "(" +
              "".join(["{:3d} ".format(sc) for sc in score[2]]) + "): " +
              "{:2d}".format(optSequenceCache[score[0]][1]) + (" * " if survive else "   ") +
              str(score[0]))
    print()

    return new_pop


def rateSequence(seq, eqs_set, dh, staggered_params=None):

    if seq not in optSequenceCache:
        optSequenceCache[seq] = [[], 0]
    cache_entry = optSequenceCache[seq]

    if cache_entry[1] > 10:
        return (cache_entry[0], [0, 0])

    print(cache_entry[1], end=" ")
    print(seq)

    start = timer()
    transformed_eqs_set = [seq.applyOpts(eqs) for eqs in eqs_set]
    end = timer()

    kernel_results = [
        bench_kernel(eqs, dh, seq.blockSize, staggered_params) for eqs in transformed_eqs_set
    ]
    kernel_registers = [k[1] for k in kernel_results]
    result = [k[0] for k in kernel_results
              ] + [k[0] * max(0.0, (len(seq.opts) - 3) * 0.1) for k in kernel_results]

    if cache_entry[1] == 0:
        cache_entry[0] = result
    else:
        for i in range(0, len(result)):
            cache_entry[0][i] = (cache_entry[0][i] * cache_entry[1] + result[i]) / (
                cache_entry[1] + 1)

    cache_entry[1] += 1

    return cache_entry[0], kernel_registers


def bench_kernel(eqs, dh, blockSize=(64, 2, 1), staggered_params=None):

    if staggered_params is None:
        kernel = ps.create_kernel(
            eqs, target="gpu", gpu_indexing_params={
                "block_size": blockSize
            }).compile()
    else:
        kernel = ps.create_staggered_kernel(
            *pack_staggered_eqs(eqs, *staggered_params),
            target="gpu",
            gpu_indexing_params={
                "block_size": blockSize
            }).compile()

    start = drv.Event()
    end = drv.Event()

    start.record()
    dh.run_kernel(kernel, timestep=1)
    dh.run_kernel(kernel, timestep=1)
    end.record()
    end.synchronize()
    msec = start.time_till(end) / 2
    return msec, kernel.num_regs
